# -*- coding: utf-8

"""
This module is designed to hold implementations of simulation models. The
implementation uses oemof/tespy.

This file is part of project dhnx (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location:

SPDX-License-Identifier: MIT
"""
import re
import warnings
from functools import partial

import networkx as nx
import numpy as np
import pandas as pd

from .graph import write_edge_data_to_graph
from .helpers import Dict
from .helpers import sum_ignore_none
from .input_output import save_results
from .model import SimulationModel

idx = pd.IndexSlice


class SimulationModelNumpy(SimulationModel):
    r"""
    Implementation of a simulation model using numpy.
    """
    def __init__(
            self, thermal_network,
            rho=971.78, c=4190, mu=0.00035, eta_pump=1, tolerance=1e-10
    ):
        super().__init__(thermal_network)
        self.results = {}

        self.nx_graph = thermal_network.to_nx_graph()

        assert nx.algorithms.tree.is_tree(self.nx_graph), \
            "Currently, only tree networks can be modeled. " \
            "Looped networks are not implemented yet."

        self.inc_mat = nx.incidence_matrix(self.nx_graph, oriented=True).todense()

        self.input_data = Dict()

        self.rho = rho  # kg/m3

        self.c = c  # J/(kg*K)

        self.mu = mu  # kg/(m*s)

        self.temp_env = thermal_network.sequences.environment.temp_env.iloc[:, 0]

        self.eta_pump = eta_pump

        self.tolerance = tolerance

        if self._concat_scalars('height') is not None:
            warnings.warn(
                "Pressure differences due to height differences are not implemented yet."
            )

    def prepare(self):

        self.prepare_hydraulic_eqn()

        self.prepare_thermal_eqn()

    def solve(self):

        self.solve_hydraulic_eqn()

        self.solve_thermal_eqn()

    def get_results(self):

        return self.results

    def prepare_hydraulic_eqn(self):
        r"""
        Prepares the input data for the hydraulic problem.
        """
        self.input_data.mass_flow = pd.DataFrame(
            0,
            columns=self.nx_graph.nodes(),
            index=self.thermal_network.timeindex
        )

        input_data = self._concat_sequences('mass_flow')

        self.input_data.mass_flow.loc[:, input_data.columns] = input_data

        self.input_data.mass_flow = self._set_producers_mass_flow(self.input_data.mass_flow)

    def prepare_thermal_eqn(self):
        r"""
        Prepares the input data for the thermal problem.
        """

        self.input_data.temp_inlet = pd.DataFrame(
            0,
            columns=self.nx_graph.nodes(),
            index=self.thermal_network.timeindex
        )

        input_data = self._concat_sequences('temp_inlet')

        self.input_data.temp_inlet.loc[:, input_data.columns] = input_data

    def solve_hydraulic_eqn(self):
        r"""
        Solves the hydraulic problem.
        """
        self.results['pipes-mass_flow'] = self._calculate_pipes_mass_flow()

        reynolds = self._calculate_reynolds()

        lamb = self._calculate_lambda(reynolds)

        pipes_dist_pressure_losses = self._calculate_pipes_distributed_pressure_losses(lamb)

        pipes_loc_pressure_losses = self._calculate_pipes_localized_pressure_losses()

        pipes_total_pressure_losses = sum_ignore_none(
            pipes_dist_pressure_losses, pipes_loc_pressure_losses
        )

        global_pressure_losses = self._calculate_global_pressure_losses(pipes_total_pressure_losses)

        pump_power = self._calculate_pump_power(global_pressure_losses)

        self.results['pipes-dist_pressure_losses'] = pipes_dist_pressure_losses

        self.results['pipes_loc_pressure_losses'] = pipes_loc_pressure_losses

        self.results['global-pressure_losses'] = global_pressure_losses

        self.results['producers-pump_power'] = pump_power

    def solve_thermal_eqn(self):
        r"""
        Solves the thermal problem.
        """
        exponent_constant = self._calculate_exponent_constant()

        temp_inlet = self._calc_temps(exponent_constant, self.input_data.temp_inlet, direction=1)

        temp_return_known = self._set_temp_return_input(temp_inlet)

        temp_return = self._calc_temps(exponent_constant, temp_return_known, direction=-1)

        pipes_heat_losses = self._calculate_pipes_heat_losses(temp_inlet) \
            + self._calculate_pipes_heat_losses(temp_return)

        global_heat_losses = pipes_heat_losses.sum(axis=1)

        global_heat_losses.name = 'global_heat_losses'

        self.results['nodes-temp_inlet'] = temp_inlet

        self.results['nodes-temp_return'] = temp_return

        self.results['pipes-heat_losses'] = pipes_heat_losses

        self.results['global-heat_losses'] = global_heat_losses

    def _concat_scalars(self, name):
        r"""
        Concatenates scalars of all components with a given variable name

        Parameters
        ----------
        name : str
            Name of the variable

        Returns
        -------
        concat_sequences : pd.DataFrame
            DataFrame containing the sequences
        """
        select_scalars = [
            scalar[name].copy().rename(index=lambda x, prefix=component: prefix + '-' + str(x))
            for component, scalar in self.thermal_network.components.items()
            if name in scalar
        ]

        if select_scalars:
            select_scalars = pd.concat(select_scalars, axis=0)

        else:
            select_scalars = None

        return select_scalars

    def _concat_sequences(self, name):
        r"""
        Concatenates sequences of all components with a given variable name

        Parameters
        ----------
        name : str
            Name of the variable

        Returns
        -------
        concat_sequences : pd.DataFrame
            DataFrame containing the sequences
        """
        select_sequences = [
            d[name].copy().rename(columns=lambda x, prefix=component: prefix + '-' + x)
            for component, d in self.thermal_network.sequences.items()
            if name in d
        ]

        concat_sequences = pd.concat(select_sequences, axis=1)

        return concat_sequences

    @staticmethod
    def _set_producers_mass_flow(m):
        r"""
        Sets the mass flow of the producer.

        Parameters
        ----------
        m : pd.DataFrame
            DataFrame with all know consumer mass flows.

        Returns
        -------
        m : pd.DataFrame
            DataFrame with all know mass flow of
            consumers and producer.
        """
        producers = [name for name in m.columns if re.search('producers', name)]

        assert len(producers) == 1, "Currently, only one producer allowed."

        m.loc[:, producers] = - m.loc[:, ~m.columns.isin(producers)].sum(1)

        return m

    def _calculate_pipes_mass_flow(self):
        r"""
        Determines the mass flow in all pipes using numpy's
        least squares function.

        Returns
        -------
        pipes_mass_flow : pd.DataFrame
            Mass flow in the pipes [kg/s]
        """

        pipes_mass_flow = {}

        for t in self.thermal_network.timeindex:

            # The order of columns in self.input_data.mass_flow fit with those of self.inc_mat
            # because the columns have been generated from the graph's nodes in
            # prepare_hydraulic_eqn()

            x, residuals, _, _ = np.linalg.lstsq(
                self.inc_mat,
                self.input_data.mass_flow.loc[t, :],
                rcond=None
            )

            assert residuals < self.tolerance, \
                f"Residuals {residuals} are larger than tolerance {self.tolerance}!"

            pipes_mass_flow.update({t: x})

        pipes_mass_flow = pd.DataFrame.from_dict(
            pipes_mass_flow,
            orient='index',
            columns=self.nx_graph.edges()
        )

        pipes_mass_flow.columns.names = ('from_node', 'to_node')

        return pipes_mass_flow

    def _calculate_reynolds(self):
        r"""
        Calculates the Reynolds number.

        .. math::

            Re = \frac{4\dot{m}}{\pi\mu D}

        Returns
        -------
        re : pd.DataFrame
            Reynolds number for every time step and pipe [-]
        """
        abs_pipes_mass_flow = self.results['pipes-mass_flow'].copy()

        abs_pipes_mass_flow = abs(abs_pipes_mass_flow)

        diameter = \
            self.thermal_network.components.pipes[['from_node', 'to_node', 'diameter']]

        diameter = 1e-3 * diameter.set_index(['from_node', 'to_node'])['diameter']

        reynolds = 4 * abs_pipes_mass_flow.divide(diameter, axis='columns') \
            / (np.pi * self.mu)

        return reynolds

    def _calculate_lambda(self, reynolds):
        r"""
        Calculates the darcy friction factor.

        .. math::

            \lambda = 0.07 \cdot Re ^ {-0.13} \cdot D ^ {-0.14}

        Parameters
        ----------
        re : pd.DataFrame
            Reynolds number for every time step and pipe [-]

        Returns
        -------
        lamb : pd.DataFrame
            Darcy friction factor for every time step and pipe [-]
        """

        factor_diameter = self.thermal_network.components.pipes[
            ['from_node', 'to_node', 'diameter']
        ]

        factor_diameter = (
            1e-3 * factor_diameter.set_index(['from_node', 'to_node'])['diameter']
        ) ** -0.14

        lamb = 0.07 * reynolds ** -0.13

        lamb = lamb.multiply(factor_diameter, axis='columns')

        return lamb

    def _calculate_pipes_distributed_pressure_losses(self, lamb):
        r"""
        Calculates the pressure losses in the pipes.

        Equal-sized inlet and return pipes are assumed which leads to
        equal mass flows and pressure losses for both. This introduces
        the initial factor of 2 in the equation.

        .. math::

            \delta p = 2 \cdot \lambda \frac{8L}{\rho \pi^2 D^5}\dot{m}^2.

        Parameters
        ----------
        lamb : pd.DataFrame
            Darcy friction factor for every time step and pipe [-]

        Returns
        -------
        pipes_pressure_losses : pd.DataFrame
            DataFrame with distributed pressure losses for inlet and return for every
            time step and pipe [Pa]
        """
        pipes_mass_flow = self.results['pipes-mass_flow'].copy()

        pipes_mass_flow_2 = pipes_mass_flow ** 2

        constant = 8 * lamb / (self.rho * np.pi**2)

        length = self.thermal_network.components.pipes[['from_node', 'to_node', 'length']]

        diameter = \
            self.thermal_network.components.pipes[['from_node', 'to_node', 'diameter']]

        length = length.set_index(['from_node', 'to_node'])['length']

        diameter = diameter.set_index(['from_node', 'to_node'])['diameter']

        diameter_5 = (1e-3 * diameter) ** 5

        pipes_pressure_losses = constant * pipes_mass_flow_2\
            .multiply(length, axis='columns')\
            .divide(diameter_5, axis='columns')

        # We multiply by the factor of two to represent the pressure losses along inlet
        # and return flow.

        pipes_pressure_losses *= 2

        return pipes_pressure_losses

    def _calculate_pipes_localized_pressure_losses(self):
        r"""
        Calculates localized pressure losses at the nodes.

        .. math::

            \Delta p_{loc} = \frac{8\zeta\dot{m}^2}{\rho \pi^2 D^4}

        Returns
        -------
        nodes_pressure_losses : pd.DataFrame
            Localized pressure losses at the nodes [Pa]
        """
        def _assign_zeta_to_pipes(flow_type):

            def func(row, level):

                res = zeta.reindex(row.index.get_level_values(level).values)

                res.index = row.index

                return res

            zeta = self._concat_scalars('zeta_' + flow_type)

            if zeta is None:
                print(f"No values for zeta_{flow_type} found. Skipping.")

                return None

            flow_direction = np.sign(self.results['pipes-mass_flow'])

            zeta_pipes = pd.DataFrame(
                columns=flow_direction.columns,
                index=flow_direction.index
            )

            zeta_pipes_forward = zeta_pipes.copy()

            zeta_pipes_backward = zeta_pipes.copy()

            func_forward = partial(func, level=0)

            zeta_pipes_forward = zeta_pipes_forward.apply(func_forward, axis=1)

            func_backward = partial(func, level=1)

            zeta_pipes_backward = zeta_pipes_backward.apply(func_backward, axis=1)

            if flow_type == 'inlet':
                zeta_pipes[flow_direction > 0] = zeta_pipes_forward

                zeta_pipes[flow_direction < 0] = zeta_pipes_backward

            elif flow_type == 'return':
                zeta_pipes[flow_direction > 0] = zeta_pipes_backward

                zeta_pipes[flow_direction < 0] = zeta_pipes_forward

            else:
                raise ValueError("Flow type has to be either inlet or return.")

            return zeta_pipes

        def _calc_loc_pressure_loss_for_flow_type(flow_type):
            zeta_pipes = _assign_zeta_to_pipes(flow_type)

            if zeta_pipes is None:
                return None

            pipes_localized_pressure_losses = {}

            for t in self.thermal_network.timeindex:
                pipes_mass_flow = self.results['pipes-mass_flow'].loc[t, :]

                mass_flow_2 = pipes_mass_flow ** 2

                mass_flow_2_over_diameter_4 = mass_flow_2.divide(diameter_4)

                mass_flow_2_over_diameter_4.name = 'mass_flow_2_over_diameter_4'

                x = constant * zeta_pipes.loc[t, :].multiply(mass_flow_2_over_diameter_4)

                pipes_localized_pressure_losses.update({t: x})

            pipes_localized_pressure_losses = pd.DataFrame.from_dict(
                pipes_localized_pressure_losses,
                orient='index'
            )

            pipes_localized_pressure_losses = pipes_localized_pressure_losses.fillna(0)

            return pipes_localized_pressure_losses

        constant = 8 / (self.rho * np.pi ** 2)

        diameter_4 = self.thermal_network.components.pipes[
            ['from_node', 'to_node', 'diameter']
        ]

        diameter_4 = diameter_4.set_index(['from_node', 'to_node'])['diameter']

        diameter_4 = (1e-3 * diameter_4) ** 4

        pipes_localized_pressure_losses_inlet = _calc_loc_pressure_loss_for_flow_type('inlet')

        pipes_localized_pressure_losses_return = _calc_loc_pressure_loss_for_flow_type('return')

        pipes_localized_pressure_losses = sum_ignore_none(
            pipes_localized_pressure_losses_inlet, pipes_localized_pressure_losses_return
        )

        return pipes_localized_pressure_losses

    def _calculate_global_pressure_losses(self, pipes_pressure_losses):
        r"""
        Calculates global pressure losses.

        Finds the path with the maximal pressure loss among from the set of
        paths from the producer to all consumers.

        Parameters
        ----------
        pipes_pressure_losses : pd.DataFrame
            Total pressure losses for every time step and pipe [Pa]

        Returns
        -------
         global_pressure_losses : pd.DataFrame
            Global pressure losses [Pa]
        """
        def calculate_path_weights(source_nodes, sink_nodes, graph, weight):

            path_weights = {}

            for sink in sink_nodes:
                for source in source_nodes:

                    path_weights[sink] = nx.dijkstra_path_length(
                        graph,
                        source=source,
                        target=sink,
                        weight=weight
                    )

            return path_weights

        def _change_graph_dir(graph, direction):
            r"""
            Takes a nx.DiGraph and returns a new graph
            with adapted edge directions according to the
            direction specified.

            Parameters
            ----------
            graph : nx.DiGraph
                Directed graph

            direction : pd.Series
                Series with the edges as index and
                1 or -1 as values, indicating the
                direction

            Returns
            -------
            dir_graph : nx.DiGraph
                Graph with adapted directions
            """
            def _get_tuple(e):
                return tuple(e[:2])

            def _swap_tuple(e):
                return tuple((*e[:2][::-1], e[2]))

            dir_edges = [
                e
                if direction.loc[_get_tuple(e)] > 0
                else _swap_tuple(e)
                for e in graph.edges(data=True)
            ]

            dir_graph = nx.DiGraph()

            dir_graph.add_edges_from(dir_edges)

            return dir_graph

        def _calculate_paths_pressure_losses():

            sink_nodes = [node for node, data in self.nx_graph.nodes(data=True) if
                          data['component_type'] == 'Consumer']

            source_nodes = [node for node, data in self.nx_graph.nodes(data=True) if
                            data['component_type'] == 'Producer']

            paths_pressure_losses = {}

            flow_direction = np.sign(self.results['pipes-mass_flow'])

            for t in self.thermal_network.timeindex:

                graph = write_edge_data_to_graph(
                    pipes_pressure_losses.loc[t, :],
                    self.nx_graph,
                    var_name='pipes_pressure_losses'
                )

                flow_dir_graph = _change_graph_dir(graph, flow_direction.loc[t, :])

                path_weights = calculate_path_weights(
                    source_nodes,
                    sink_nodes,
                    flow_dir_graph,
                    'pipes_pressure_losses'
                )

                paths_pressure_losses[t] = path_weights

            paths_pressure_losses = \
                pd.DataFrame.from_dict(paths_pressure_losses, orient='index')

            return paths_pressure_losses

        paths_pressure_losses = _calculate_paths_pressure_losses()

        # Here, we take the path with the maximum pressure losses and assume that the other
        # consumer's valves are adjusted so that in sum, the pressure losses along all paths are
        # equal.

        global_pressure_losses = paths_pressure_losses.max(axis=1)

        return global_pressure_losses

    def _calculate_pump_power(self, global_pressure_losses):
        r"""
        Calculates the pump power.

        .. math::

            P_{el. pump} = \frac{1}{\eta_{el}\eta_{hyd}}\frac{\Delta p }{\rho} \dot{m}

        Parameters
        ----------
        global_pressure_losses : pd.DataFrame
            Global pressure losses [Pa]

        Returns
        -------
         pump_power : pd.Series
            Pump power [W]
        """
        producers = [
            node for node, data in self.nx_graph.nodes(data=True)
            if data['component_type'] == 'Producer'
        ]

        mass_flow_producers = self.results['pipes-mass_flow'].iloc[
            :, self.results['pipes-mass_flow'].columns.isin(
                producers, level=0)].sum(axis=1)

        pump_power = mass_flow_producers * global_pressure_losses / (self.eta_pump * self.rho)

        return pump_power

    def _calculate_exponent_constant(self):
        r"""
        Calculates the constant part of the exponent that determines the
        cooling of the medium in the pipes.

        .. math::

            exp_{const} = - \frac{U \pi D L }{c}

        Returns
        -------
        exponent_constant : np.matrix
            Constant part of the exponent [kg/s]
        """

        heat_transfer_coefficient = nx.adjacency_matrix(
            self.nx_graph, weight='heat_transfer_coefficient').todense()

        diameter = 1e-3 * nx.adjacency_matrix(self.nx_graph, weight='diameter').todense()

        length = nx.adjacency_matrix(self.nx_graph, weight='length').todense()

        exponent_constant = - np.pi \
            * np.multiply(heat_transfer_coefficient, np.multiply(diameter, length)) \
            / self.c

        return exponent_constant

    def _calc_temps(self, exponent_constant, known_temp, direction):
        r"""
        Calculate temperatures

        .. math::

            T_{out} = T_{env} + (T_{in} - T_{env}) \cdot exp\{exp_{const} \cdot exp_{var}\} =
            T_{out} = T_{env} + (T_{in} - T_{env}) \cdot exp\{-\frac{U \pi D L}{c \cdot \dot{m}}\}

        Parameters
        ----------
        exponent_constant : np.array
            Constant part of the exponent [kg/s]

        known_temp : pd.DataFrame
            Known temperatures at producers or consumers [°C]

        direction : +1 or -1
            For inlet and return flow [-]

        Returns
        -------
        temp_df : pd.DataFrame
            DataFrame containing temperatures for all nodes [°C]
        """
        # TODO: Rethink function layout and naming

        temps = {}

        for t in self.thermal_network.timeindex:

            # Divide exponent matrix by current pipes-mass_flows.
            data = self.results['pipes-mass_flow'].loc[t, :].copy()

            data = 1 / data

            graph_with_data = write_edge_data_to_graph(
                data, self.nx_graph, var_name='pipes-mass_flow'
            )

            inverse_mass_flows = nx.adjacency_matrix(
                graph_with_data, weight='pipes-mass_flow'
            ).todense()

            exponent = np.multiply(exponent_constant, inverse_mass_flows)

            matrix = np.exp(exponent)

            # Clear out elements where matrix was zero before exponentiation. This could be
            # replaced by properly passing `where` to np.multiply in the line above.
            matrix = np.multiply(matrix, nx.adjacency_matrix(self.nx_graph).todense())

            # Adapt matrix
            if direction == 1:
                matrix = matrix.T
                normalisation = np.array(nx.adjacency_matrix(self.nx_graph).sum(0)).flatten()

                normalisation = \
                    np.divide(np.array([1]), normalisation, where=normalisation != 0)

            elif direction == -1:
                normalisation = np.array(nx.adjacency_matrix(self.nx_graph).sum(1)).flatten()

                normalisation = \
                    np.divide(np.array([1]), normalisation, where=normalisation != 0)

            else:
                raise ValueError("Direction has to be either 1 or -1.")

            normalisation = np.diag(normalisation)

            matrix = np.dot(normalisation, matrix)

            matrix = np.identity(len(matrix)) - matrix

            vector = np.array(known_temp.loc[t])

            vector[vector != 0] -= self.temp_env.loc[t]

            x, _, _, _ = np.linalg.lstsq(
                matrix,
                vector,
                rcond=None
            )

            temps.update({t: x + self.temp_env.loc[t]})

        temp_df = pd.DataFrame.from_dict(
            temps,
            orient='index',
            columns=self.nx_graph.nodes()
        )

        return temp_df

    def _set_temp_return_input(self, temp_inlet):
        r"""
        Sets the temperature of the return pipes
        at the consumers.

        .. math::

            T_{cons,r} = T_{cons,i} - T_{cons,drop}

        Parameters
        ----------
        temp_inlet : pd.DataFrame
            Known inlet temperature [°C]

        Returns
        -------
        temp_return : pd.DataFrame
            Return temperature with the consumers values set [°C]
        """

        temp_return = pd.DataFrame(
            0,
            columns=self.nx_graph.nodes(),
            index=self.thermal_network.timeindex
        )

        temp_drop = self._concat_sequences('delta_temp_drop')

        temp_return.loc[:, temp_drop.columns] = temp_inlet.loc[:, temp_drop.columns] - temp_drop

        return temp_return

    def _calculate_pipes_heat_losses(self, temp_node):
        r"""
        Calculates the pipes' heat losses given the
        temperatures.

        .. math::

            \dot{Q}_{losses} = c \cdot \dot{m} \cdot \Delta T

        Parameters
        ----------
        temp_node : pd.DataFrame
            Temperatures at the nodes [°C]

        Returns
        -------
        pipes_heat_losses : pd.DataFrame
            Heat losses in the pipes [W]
        """

        pipes_heat_losses = {}

        for i, row in temp_node.iterrows():

            mass_flow = self.results['pipes-mass_flow'].loc[i, :].copy()

            temp_difference = np.abs(np.array(np.dot(row, self.inc_mat)).flatten())

            pipes_heat_losses[i] = self.c * mass_flow.multiply(temp_difference, axis=0)

        pipes_heat_losses = pd.DataFrame.from_dict(pipes_heat_losses, orient='index')

        return pipes_heat_losses


def simulate(thermal_network, results_dir=None):
    r"""
    Takes a thermal network and returns the result of
    the simulation.

    Parameters
    ----------
    thermal_network

    Returns
    -------
    results : dict
    """
    model = SimulationModelNumpy(thermal_network)

    model.prepare()

    model.solve()

    results = model.get_results()

    if results_dir is not None:
        save_results(results, results_dir)

    return results



# -*- coding: utf-8 -*-

"""
This modules holds functions for geometry operations, that are needed for
the geometry processing module `connect_points.py`.

This file is part of project dhnx (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location: https://github.com/oemof/DHNx

This module is not fully tested yet, so use it with care.

SPDX-License-Identifier: MIT
"""

try:
    import geopandas as gpd

except ImportError:
    print("Need to install geopandas to process geometry data.")

try:
    import shapely
    from shapely import wkt
    from shapely.geometry import LineString
    from shapely.geometry import MultiLineString
    from shapely.geometry import Point
    from shapely.geometry import mapping
    from shapely.ops import linemerge
    from shapely.ops import nearest_points
    from shapely.ops import unary_union
except ImportError:
    print("Need to install shapely to process geometry.")

import logging

import matplotlib.pyplot as plt
import pandas as pd

logger = logging.getLogger(__name__)  # Create a logger for this module


def create_forks(lines):
    """
    Creates a forks(nodes) GeoDataFrame from a "line"-GeoDataFrame
    based on the end-points of each LineString.

    Also, an index for every fork is given, and the columns 'full-id' (="forks-" + index"),
    'lat' and 'lon', which results from the geometry, are added to the GeoDataFrame.

    Parameters
    ----------
    lines : geopandas.GeoDataFrame

    Returns
    -------
    geopandas.GeoDataFrame : GeoDataFrame with Points as geometry.

    """
    nodes = gpd.GeoDataFrame(geometry=[], crs=lines.crs)

    for _, j in lines.iterrows():
        geom = j['geometry']
        p_0 = Point(geom.boundary.geoms[0])
        p_1 = Point(geom.boundary.geoms[-1])
        nodes = pd.concat(
            [nodes, gpd.GeoDataFrame(geometry=[p_0, p_1], crs=lines.crs)],
            ignore_index=True)

    # transform geometry into wkt
    nodes["geometry_wkt"] = nodes["geometry"].apply(lambda geom: geom.wkt)

    # drop duplicates of geometry column
    nodes = nodes.drop_duplicates(["geometry_wkt"])

    # create shapely geometry again
    nodes["geometry"] = nodes["geometry_wkt"].apply(
        lambda geom: wkt.loads(geom))  # pylint: disable=unnecessary-lambda

    # set index for forks
    nodes = nodes.reset_index(drop=True)
    nodes['id'] = nodes.index
    nodes['id_full'] = 'forks-' + nodes['id'].apply(str)
    nodes['lat'] = nodes['geometry'].apply(lambda x: x.y)
    nodes['lon'] = nodes['geometry'].apply(lambda x: x.x)
    nodes.set_index('id', drop=True, inplace=True)

    return nodes


def insert_node_ids(lines, nodes):
    """Create the columns `from_node`, `to_node` and insert the node ids.

    The node ids are called e.g. forks-3, consumers-5 etc.
    The updated "line"-GeoDataFrame is returned.

    Parameters
    ----------
    lines : geopandas.GeoDataFrame
    nodes : geopandas.GeoDataFrame

    Returns
    -------
    geopandas.GeoDataFrame
    """
    nodes['geo_wkt'] = nodes.geometry.apply(
        lambda x: wkt.dumps(x, output_dimension=2))
    nodes.set_index('geo_wkt', drop=True, inplace=True)

    # add id to gdf_lines for starting and ending node point as wkt
    lines['b0_wkt'] = lines.geometry.apply(
        lambda geom: wkt.dumps(geom.boundary.geoms[0], output_dimension=2))
    lines['b1_wkt'] = lines.geometry.apply(
        lambda geom: wkt.dumps(geom.boundary.geoms[-1], output_dimension=2))

    def match_multipoint(point_wkt):
        """Return id_full from matching 'nodes' for each point in point_wkt.

        This is necessary if point_wkt contains MultiPoint objects, which
        result from multiple connection lines, and not only single Point
        objects.
        """
        point_line = wkt.loads(point_wkt)
        for point_node in nodes.index:
            if point_line.within(wkt.loads(point_node)):
                return nodes.loc[point_node, 'id_full']
        logger.error("Point not found: %s", point_wkt)
        return False

    try:
        lines['from_node'] = lines['b0_wkt'].apply(
            lambda x: nodes.at[x, 'id_full'])
        lines['to_node'] = lines['b1_wkt'].apply(lambda x: match_multipoint(x))
    except KeyError as e:
        errors = ([wkt.loads(x) for x in lines['b0_wkt']
                   if x not in nodes['id_full']])
        errors.extend([wkt.loads(x) for x in lines['b1_wkt']
                       if not match_multipoint(x)])
        gdf_errors = gpd.GeoDataFrame(geometry=errors, crs=lines.crs)
        ax = lines.plot()
        gdf_errors.plot(ax=ax, color='red', label='Point(s) causing error')
        plt.legend()
        plt.show()
        try:
            nodes.to_file('debug_nodes.geojson')
            gdf_errors.to_file('debug_error_points.geojson')
            # lines[[lines.geometry.name, 'type', 'b0_wkt', 'b1_wkt']
            #       ].to_file('debug_lines.geojson')
        except Exception as e:
            breakpoint()
            logger.error(e)

        raise KeyError("This error indicates specific problems with the data. "
                       "A plot of the problematic point(s) is shown.") from e

    lines.drop(axis=1, inplace=True, labels=['b0_wkt', 'b1_wkt'])

    return lines


def check_double_points(gdf, radius=0.001, id_column=None):
    """Check for points, which are close to each other.

    In case, two points are close, the index of the points are printed.

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame
        GeoDataFrame with Points as geometry.
    radius : float
        Maximum distance.
    id_column : str or None
        Column name which should be printed in case of near points.
        If None, the index is printed.

    Returns
    -------
    list : Indices of "near" points.
    """

    l_ids = []
    count = 0

    for r, c in gdf.iterrows():

        point = c['geometry']
        gdf_other = gdf.drop([r])
        # Prevent OSError, see https://github.com/oemof/DHNx/issues/107
        other_points = unary_union(list(gdf_other['geometry']))

        # x1 = nearest_points(point, other_points)[0]
        x2 = nearest_points(point, other_points)[1]

        if point.distance(x2) <= radius:
            l_ids.append(r)

            if id_column is None:
                print_name = r
            else:
                print_name = c[id_column]

            logger.info(
                'Node {} has a near neighbour! '
                'Distance {}'.format(print_name, point.distance(x2))
            )

            count += 1

    if count > 0:
        logger.info('Number of duplicated points: %s', count)
    else:
        logger.info(
            'Check passed: No points with a distance closer than {}'.format(radius))

    return l_ids


def gdf_to_df(gdf):
    """Converts a GeoDataFrame to a pandas.DataFrame by deleting the geometry column."""

    df = pd.DataFrame(
        gdf[[col for col in gdf.columns if col != 'geometry']])

    return df


def split_multilinestr_to_linestr(gdf_input):
    """Simplifies GeoDataFrames with LineStrings as geometry.

    The LineStrings (whether LineStrings, or MulitLineStings) are split into
    LineStrings with only two coordinates, one starting and one ending point.

    The other values of the GeoDataFrame are copied to the new rows
    for each row, who's geometry is split.

    Parameters
    ----------
    gdf_lines : geopandas.GeoDataFrame

    Returns
    -------
    geopandas.GeoDataFrame
    """
    gdf_lines = gdf_input.copy()

    new_lines = gpd.GeoDataFrame()

    # first: split MultiLineString into LineStrings
    for i, b in gdf_lines.iterrows():

        geom = b['geometry']

        if geom.geom_type == 'MultiLineString':

            multilinestrings = []

            for line in mapping(geom)["coordinates"]:
                multilinestrings.append(LineString(line))

            for multiline in multilinestrings:
                new_row = b.copy()
                new_row['geometry'] = multiline
                new_lines = pd.concat([new_lines, new_row.to_frame().T],
                                      ignore_index=True, sort=False)

            gdf_lines.drop(index=i, inplace=True)

    gdf_lines = pd.concat([gdf_lines, new_lines], ignore_index=True,
                          sort=False)

    gdf_lines['geometry'].crs = gdf_input.crs

    # second: split LineStrings into single Linestrings
    new_lines = gpd.GeoDataFrame()
    for i, b in gdf_lines.iterrows():

        geom = b['geometry']

        if len(geom.coords) > 2:

            num_new_lines = len(geom.coords) - 1

            for num in range(num_new_lines):
                new_row = b.copy()
                new_row['geometry'] = \
                    LineString([geom.coords[num], geom.coords[num + 1]])
                new_lines = pd.concat([new_lines, new_row.to_frame().T],
                                      ignore_index=True, sort=False)

            gdf_lines.drop(index=i, inplace=True)

    gdf_lines = pd.concat([gdf_lines, new_lines], ignore_index=True,
                          sort=False)

    gdf_lines['geometry'].crs = gdf_input.crs

    return gdf_lines


def weld_segments(gdf_line_net, gdf_line_gen, gdf_line_houses,
                  debug_plotting=False):
    """Weld continuous line segments together and cut loose ends.

    This is a public function that recursively calls the internal function
    _weld_segments(), until the problem cannot be simplified further.

    Find all lines that only connect to one other line and connect those
    to a single MultiLine object. Points that connect to Generators and
    Houses are not simplified. Loose ends are shortened where possible.

    Parameters
    ----------
    gdf_line_net : GeoDataFrame
        Potential pipe network.
    gdf_line_gen : GeoDataFrame
        Generators that need to be connected.
    gdf_line_houses : GeoDataFrame
        Houses that need to be connected.
    debug_plotting : bool, optional
        Plot the selection process.

    Returns
    -------
    gdf_line_net_new : GeoDataFrame
        Simplified potential pipe network.

    """
    gdf_line_net_last = gdf_line_net
    gdf_line_net_new = _weld_segments(gdf_line_net, gdf_line_gen,
                                      gdf_line_houses, debug_plotting)
    # Now do all of this recursively
    while len(gdf_line_net_new) < len(gdf_line_net_last):
        logger.info('Welding lines... reduced from {} to {} lines'.format(
            len(gdf_line_net_last), len(gdf_line_net_new)))
        gdf_line_net_last = gdf_line_net_new
        gdf_line_net_new = _weld_segments(gdf_line_net_new, gdf_line_gen,
                                          gdf_line_houses, debug_plotting)
    logger.info('Welding lines... done')
    return gdf_line_net_new


def _weld_segments(gdf_line_net, gdf_line_gen, gdf_line_houses,
                   debug_plotting=False):
    """Weld continuous line segments together and cut loose ends.

    Find all lines that only connect to one other line and connect those
    to a single MultiLine object. Points that connect to Generators and
    Houses are not simplified. Loose ends are shortened where possible.

    Parameters
    ----------
    gdf_line_net : GeoDataFrame
        Potential pipe network.
    gdf_line_gen : GeoDataFrame
        Generators that need to be connected.
    gdf_line_houses : GeoDataFrame
        Houses that need to be connected.
    debug_plotting : bool, optional
        Plot the selection process.

    Returns
    -------
    gdf_line_net_new : GeoDataFrame
        Simplified potential pipe network.

    """
    crs = gdf_line_net.crs
    gdf_line_net_new = gpd.GeoDataFrame(geometry=[], crs=crs)
    gdf_merged_all = gpd.GeoDataFrame(geometry=[], crs=crs)
    gdf_deleted = gpd.GeoDataFrame(geometry=[], crs=crs)
    # Merge generator and houses line DataFrames to 'external' lines
    gdf_line_ext = pd.concat([gdf_line_gen, gdf_line_houses])

    for _, b in gdf_line_net.iterrows():
        def debug_plot(neighbours, color='red'):
            """Plot base map, current segment (with color) and neighbours."""
            if debug_plotting:
                _, ax = plt.subplots(1, 1, dpi=300)
                gdf_line_net.plot(ax=ax, color='blue')
                gdf_line_ext.plot(ax=ax, color='green')
                if len(neighbours) > 0:  # Prevent empty plot warning
                    neighbours.plot(ax=ax, color='orange')
                gpd.GeoDataFrame(geometry=[geom]).plot(ax=ax, color=color)

        geom = b.geometry  # The current line segment
        gdf_b = gpd.GeoDataFrame(b.to_frame().T, crs=crs)

        if any(gdf_merged_all.geometry.contains(geom)):
            # Drop this object, because it is contained within a merged object
            continue  # Continue with the next line segment

        if geom.is_ring:
            # Drop this object, because rings are not valid options for street
            # segments, since they have no start or end
            continue  # Continue with the next line segment

        # Find all neighbours of the current segment.
        # Neighbours are geometries whose boundary "touches" the current
        # geometry. We need to use the boundary to allow roads that cross
        # themselves, e.g. for spiral ramps. Also they must not be equal
        # to the current segment.
        neighbours = gdf_line_net[
            (gdf_line_net.geometry.boundary.touches(geom)
             & ~gdf_line_net.geometry.geom_equals(geom)
             )]
        # If all of the neighbours touch each other, it is the
        # last segment before an intersection, which can be removed.
        # The tests needs to be "touches" OR "equals", since per definition
        # a line geometry cannot "touch" itself
        if all([all(neighbours.geometry.touches(neighbour)
                    | neighbours.geometry.geom_equals(neighbour))
                for neighbour in neighbours.geometry]):
            # Treat as if there was only one neighbour (like end segment)
            neighbours = neighbours.head(1)

        if len(neighbours) <= 1:
            # This is a potentially unused end segment
            unused = True

            # Test if one end touches an 'external' line, while the other
            # end touches a network line segment
            p1 = geom.boundary.geoms[0]
            p2 = geom.boundary.geoms[-1]
            p1_neighbours = neighbours.geometry.intersects(p1).to_list()
            p2_neighbours = neighbours.geometry.intersects(p2).to_list()

            if (any(gdf_line_ext.geometry.touches(p1))
               and p2_neighbours.count(True) > 0):
                unused = False
            elif (any(gdf_line_ext.geometry.touches(p2))
                  and p1_neighbours.count(True) > 0):
                unused = False
            elif (any(gdf_line_ext.geometry.touches(geom))
                  and not any(gdf_line_net.geometry.touches(geom))):
                # The current segment is touched by an external line, but not
                # by any other network segment. Select connected external line
                geom_ext = unary_union(gdf_line_ext[
                    gdf_line_ext.geometry.touches(geom)].geometry)
                # Test if the network line touched by the external line
                # is equal to the current line segement, i.e. the external
                # line is not connected to any other network line
                if gdf_line_net[gdf_line_net.geometry.touches(geom_ext)
                                ].geometry.geom_equals(geom).all():
                    logger.warning(
                        "Welding is about to remove a street network line "
                        "segment that is connected to nothing but a building "
                        "connection line. This would leave the building "
                        "unconnected, so the segment is not removed. "
                        "This indicates a bug in the welding logic or an "
                        "issue in the input data, e.g. an initial street "
                        "network where not all lines are connected. If the "
                        "remaining process fails, this may be the cause.")
                    unused = False  # Keep line, despite not being useful

            if unused:
                # If truly unused, we can discard it to simplify the network
                debug_plot(neighbours, color='white')
                gdf_deleted = pd.concat([gdf_deleted, gdf_b],
                                        ignore_index=True)
            else:
                # Keep it, if it touches a generator or a house
                debug_plot(neighbours, color='black')
                gdf_line_net_new = pd.concat([gdf_line_net_new, gdf_b],
                                             ignore_index=True)
            continue  # Continue with the next line segment

        if len(neighbours) > 2:
            # This segment has more than two neighbours. This means it is
            # part of an intersection, which we do not simplify futher.
            # However, we can check if either endpoint of the current segment
            # only has one neighbour. Then that one can still be merged.
            p1 = geom.boundary.geoms[0]
            p2 = geom.boundary.geoms[-1]
            p1_neighbours = neighbours.geometry.intersects(p1).to_list()
            p2_neighbours = neighbours.geometry.intersects(p2).to_list()
            if p1_neighbours.count(True) == 1:  # Only one neighbour allowed
                neighbours = neighbours[p1_neighbours]  # Neighbour to merge
            elif p2_neighbours.count(True) == 1:  # Only one neighbour allowed
                neighbours = neighbours[p2_neighbours]  # Neighbour to merge
            else:  # Keep this segment. Multiple lines meet at an intersection
                gdf_line_net_new = pd.concat([gdf_line_net_new, gdf_b],
                                             ignore_index=True)
                debug_plot(neighbours, color='green')
                continue  # Continue with the next line segment

        if len(neighbours) == 2:
            # There are excactly two separate neighbours that can be merged
            pass  # Run the rest of the loop

        # Before merging, we need to further clean up the list of neighbours
        neighbours_list = []
        for neighbour in neighbours.geometry:
            if any(gdf_deleted.geometry.geom_equals(neighbour)):
                continue  # Do not use neighbour that has already been deleted
            if any(gdf_line_net_new.geometry.contains(neighbour)):
                continue  # Prevent creating dublicates
            if any(gdf_line_ext.geometry.intersects(neighbour)):
                mask = gdf_line_ext.geometry.intersects(neighbour)
                houses = gdf_line_ext[mask]
                # Neighbour intersects with external, but geom does not
                if all(houses.geometry.disjoint(geom)):
                    neighbours_list.append(neighbour)
                else:  # No not merge neighbour intersecting with external
                    continue
            elif any(neighbours.geometry.touches(neighbour)):
                neighbours_list = []  # The two neighbours touch
                break  # This is a intersection that cannot be simplified
            else:  # Choose neighbour for merging
                neighbours_list.append(neighbour)
        neighbours = gpd.GeoDataFrame(geometry=neighbours_list, crs=crs)

        if len(neighbours) == 0:
            # If no neighbours are left now, continue with next line segment
            gdf_line_net_new = pd.concat([gdf_line_net_new, gdf_b],
                                         ignore_index=True)
            continue

        # Create list of all elements that should be merged
        lines = [geom] + list(neighbours.geometry)
        try:  # Works when all elements are LineStrings
            # Combine lines into a multi-linestring
            multi_line = MultiLineString(lines)
        except NotImplementedError:  # Fails if there is a MultiLineString
            lines_ = []  # Create a new list of lines, without MultiLineStrings
            for line in lines:
                if line.type == 'MultiLineString':
                    lines_ += list(line)  # Split the MultiLineString
                else:  # Linestring
                    lines_.append(line)
            # Now combine all of those into MultiLineString
            multi_line = MultiLineString(lines_)

        # Merge the MultiLineString into a single object
        merged_line = linemerge(multi_line)
        gdf_merged = gpd.GeoDataFrame(geometry=[merged_line], crs=crs)
        debug_plot(neighbours)  # Plot the segments before the merge
        debug_plot(gdf_merged, color='orange')  # ...and after the merge
        gdf_line_net_new = pd.concat([gdf_line_net_new, gdf_merged],
                                     ignore_index=True)
        gdf_merged_all = pd.concat([gdf_merged_all, gdf_merged],
                                   ignore_index=True)

    return gdf_line_net_new


def drop_parallel_lines(gdf):
    """Keep only the shortest of all lines connecting the same two points.

    This prevents an error in eomof.solph that will occur if multiple lines
    connect the same two points in a network.

    These can be two actually distinct paths between two points, or two
    identical lines on top of each other. When downloading streets with osmnx,
    this can introduce such duplicates where the two have the attributes
    'reversed=True' and 'reversed=False'

    This function modifies the GeoDataFrame in place and resets the index.
    """
    # Stores each LineString's endpoints and length in temporary columns
    gdf['5d7u6j_endpoints'] = gdf.geometry.apply(
        lambda line: tuple(sorted([line.coords[0], line.coords[-1]])))
    gdf['5d7u6j_length'] = gdf.geometry.length

    # Group by endpoints and keep only the shortest LineString for each group
    gdf = (gdf.sort_values('5d7u6j_length')
           .groupby('5d7u6j_endpoints').first()
           .set_crs(gdf.crs)  # The groupby operation removes crs info
           .reset_index(drop=True)  # Drop the temporary columns
           .drop(columns=['5d7u6j_length'])  # Drop the temporary columns
           )
    return gdf


def check_crs(gdf, crs=4647, force_2d=True):
    """Convert CRS to EPSG:4647 - ETRS89 / UTM zone 32N (zE-N).

    This is the (only?) Coordinate Reference System that gives the correct
    results for distance calculations.

    Note about enforcing 2D geometries:

    Most underlying functions (e.g. distance measurement in shapely) do
    not support 3D geometries. Having z-coordinates in the
    data can cause issues in several places in the code, especially when
    e.g. buildings contain z coordinates, but streets do not.

    The savest option currently is to force 2d on all input geometries.

    Parameters
    ----------
    gdf : GeoDataFrame
        The GeoDataFrame to update.
    crs : int (optional)
        EPSG code for a coordinate reference system to convert to.
        Default is 4647.
    force_2d : boolean (optional)
        If True, enforce reduction of 3D geometry to 2D. Default is True.

    Returns
    -------
    gdf : GeoDataFrame
        Updated GeoDataFrame.


    """
    if gdf.crs.to_epsg() != crs:
        gdf.to_crs(epsg=crs, inplace=True)
        logger.info('CRS of GeoDataFrame converted to EPSG:{0}'.format(crs))

    if force_2d and gdf.has_z.any():
        logger.debug("Reducing 3D geometry to 2D for compatibility")
        # Alternatively, use "gdf.force_2d()" with geopandas>0.14.3
        gdf.geometry = shapely.force_2d(gdf.geometry)  # requires shapely>=2.0

    return gdf



# -*- coding: utf-8

"""
This script calculates the heat transfer at the consumer
for a simple tree network.
"""
import math
import os

import numpy as np
import pandas as pd

path_file = os.path.dirname(__file__)
path = os.path.abspath(os.path.join(path_file, os.pardir, os.pardir, os.pardir))
input_data = os.path.join(path, "examples", "simulation", "tree")
result_path = os.path.join(path_file, "sequences")

if not os.path.exists(result_path):
    os.mkdir(result_path)


def read_data(input_value):
    r"""
    This function is reading the data of a csv with a name given as input value
    """
    return pd.read_csv(os.path.join(input_data, input_value + ".csv"), index_col=0)


# Read input data for every csv component
pipes = read_data("pipes")
temp_drop = pd.read_csv(input_data + "/sequences/consumers-delta_temp_drop.csv")
mass_flow = pd.read_csv(input_data + "/sequences/consumers-mass_flow.csv")

# Constants for calculation
t_env = 10  # [°C]
t_prod_i = pd.DataFrame(data={"t_prod_i": 130 * np.ones(len(mass_flow))})  # [°C]
c = 4190  # [J/kg*K]
pi = math.pi

# Initialize variables of type dataframe (needed for later calculations)
U_spec, t_cons_i, t_cons_r, t_fork_r, Q_loss_i, Q_loss_r, Q_cons, Q_loss_glob = [
    pd.DataFrame() for variable in range(8)
]

# Adjust mass flows and temp drop to a dataframe containing all data in correct order
# Get mass flows of all consumers
mass_flow_total = mass_flow.iloc[:, 1:]
# Rename the columns to pipes naming convention
mass_flow_total.columns = ["1", "2"]
# Calculate producer mass flow as sum of consumer mass flows
mass_flow_total["0"] = mass_flow_total["1"] + mass_flow_total["2"]
# Change order of columns for later calculation
mass_flow_total = mass_flow_total[["0", "1", "2"]]

# Get temperature drop of all consumers
temp_drop = temp_drop.iloc[:, 1:]
# Rename the columns to pipes naming convention
temp_drop = temp_drop.rename(columns={"0": "1", "1": "2"})


def calc_temp_heat_loss(t_in, pos):
    r"""
    This function calculates the pipe's outlet temperature
    out of the inlet temperature due to heat losses

    Parameters
    ----------
    t_in : Series
           Temperature entering the pipe
    pos : int
          Position of node

    Returns
    -------
    t_out : Series
            Temperature leaving the pipe
    """
    t_out = t_env + (t_in - t_env) * np.exp(
        -pipes["heat_transfer_coefficient"].iloc[pos]
        * pi
        * pipes["diameter"].iloc[pos]
        / 1000
        * pipes["length"].iloc[pos]
        / (c * mass_flow_total[str(pos)])
    )
    return t_out


def calc_heat_loss(m, t_in, t_out):
    r"""
    This function calculates heat losses

    Needs to be adapted in case heat capacity is not constant as assumed

    Parameters
    ----------
    m : Series
        Mass flow [kg/s]
    t_in : Series
           Inlet temperature [K]
    t_out : Series
            Outlet temperature [K]

    Returns
    -------
    Heat flow [W]
    """
    return m * c * (t_in - t_out)


# Calculate inlet temperature at fork
t_fork_i = pd.DataFrame(data={"0": calc_temp_heat_loss(t_prod_i["t_prod_i"], 0)})

# Calculate heat loss at pipe from producer to fork
Q_loss_i["0"] = calc_heat_loss(
    mass_flow_total["0"], t_prod_i["t_prod_i"], t_fork_i["0"]
)

for index in list(temp_drop):
    # Calculate inlet temperature at consumers
    t_cons_i[index] = calc_temp_heat_loss(t_fork_i["0"], int(index))
    # Calculate return temperature at consumers
    t_cons_r[index] = t_cons_i[index] - temp_drop[index]
    # Calculate return temperature at fork
    t_fork_r[index] = calc_temp_heat_loss(t_cons_r[index], int(index))
    # Calculate heat losses at pipe from fork to consumer
    Q_loss_i[index] = calc_heat_loss(
        mass_flow_total[index], t_fork_i["0"], t_cons_i[index]
    )
    # Calculate heat losses at pipe from consumer to fork
    Q_loss_r[index] = calc_heat_loss(
        mass_flow_total[index], t_cons_r[index], t_fork_r[index]
    )
    # Calculate heat transfer at consumers with temperature drop
    Q_cons[index] = calc_heat_loss(
        mass_flow_total[index], t_cons_i[index], t_cons_r[index]
    )

# Calculate temperature of mixture at fork return
# Note with these input values irrelevant
# because the temperatures coming from the consumers are the same.
# Needs to be adapted in case capacity is not constant as assumed
t_fork_r_mix = pd.DataFrame(
    data={
        "0": (
            mass_flow_total["1"] * t_fork_r["1"] + mass_flow_total["2"] * t_fork_r["2"]
        )
        / mass_flow_total["0"]
    }
)

# Calculate return temperature at producer
t_prod_r = pd.DataFrame(data={"0": calc_temp_heat_loss(t_fork_r_mix["0"], int("0"))})

# Calculate inlet temperature of nodes
t_nodes_i = pd.DataFrame(
    data={
        "producers-0": t_prod_i["t_prod_i"],
        "forks-0": t_fork_i["0"],
        "consumers-0": t_cons_i["1"],
        "consumers-1": t_cons_i["2"],
    }
)

# Calculate return temperature of nodes
t_nodes_r = pd.DataFrame(
    data={
        "producers-0": t_prod_r["0"],
        "forks-0": t_fork_r_mix["0"],
        "consumers-0": t_cons_r["1"],
        "consumers-1": t_cons_r["2"],
    }
)

# Calculate heat loss at pipe from fork to producer
Q_loss_r["0"] = calc_heat_loss(mass_flow_total["0"], t_fork_r_mix["0"], t_prod_r["0"])

# Calculate total heat losses (inlet and return)
Q_loss = Q_loss_i + Q_loss_r

# Calculate global heat losses
Q_loss_glob = pd.DataFrame(data={"global_heat_losses": np.zeros(len(mass_flow_total))})
for index, node in enumerate(mass_flow_total):
    Q_loss_glob["global_heat_losses"] = (
        Q_loss_glob["global_heat_losses"] + Q_loss[str(index)]
    )


# Print results
def parameters():
    r"""
    Writes results in Dictionary

    Returns
    -------
    parameter : dict
                Dictionary with results
    """
    param_dict = {
        "Mass flow [kg/s]": mass_flow_total,
        "Inlet temperature at producer T_prod_i [°C]": t_prod_i,
        "Return temperature at producer T_prod_r [°C]": t_prod_r,
        "Inlet temperature at fork T_fork_i [°C]": t_fork_i,
        "Return temperature at fork T_fork_r [°C]": t_fork_r_mix,
        "Inlet temperature at consumer T_c_i [°C]": t_cons_i,
        "Return temperature at consumer T_c_out [°C]": t_cons_r,
        "Inlet temperature nodes T_nodes_i [°C]": t_nodes_i,
        "Return temperature nodes T_nodes_r [°C]": t_nodes_r,
        "Heat losses Q_loss [W]": Q_loss,
        "Global heat losses Q_loss_glob [W]": Q_loss_glob,
        "Heat transfer at consumers Q [W]": Q_cons,
    }
    return param_dict


parameter = parameters()


def print_parameters():
    r"""
    Prints the parameters
    """
    dash = "-" * 60

    print("\n" + dash)
    print("Results at producer (0), consumer 1 (1) and consumer 2 (2)")
    print(dash)

    for name, param in parameter.items():
        print(name + "\n" + str(param) + "\n\n")

    print(dash)


print_parameters()


# Save results to csv
for value, params in parameter.items():
    params.insert(0, "snapshot", np.arange(len(mass_flow_total)))


result_name = [
    "pipes-mass_flow.csv",
    "nodes-temp_inlet.csv",
    "nodes-temp_return.csv",
    "pipes-heat_losses.csv",
    "global-heat_losses.csv",
]

result_list = [list(parameter.keys())[0]] + list(parameter.keys())[7:11]

for index, value in enumerate(result_list):
    parameter[value].to_csv(os.path.join(result_path, result_name[index]), index=False)



from . import add_components  # noqa: F401
from . import dhs_nodes  # noqa: F401
from . import oemof_heatpipe  # noqa: F401
from . import optimization_models  # noqa: F401
from . import precalc_hydraulic  # noqa: F401



import matplotlib.pyplot as plt

from dhnx.plotting import InteractiveMap
from dhnx.plotting import StaticMap
from dhnx.network import ThermalNetwork

# initialize a thermal network
thermal_network = ThermalNetwork()

# load data from csv
thermal_network.from_csv_folder('data_csv_input')

# save thermal network to csv
thermal_network.to_csv_folder('data_csv_output')

# get graph of thermal network
graph = thermal_network.to_nx_graph()

# plot static map
static_map = StaticMap(thermal_network)

static_map.draw(background_map=False)
plt.savefig('static_map_wo_background.png')

static_map.draw(background_map=True)
plt.savefig('static_map_w_background.png')

# plot interactive map
interactive_map = InteractiveMap(thermal_network)
map = interactive_map.draw()
map.save('interactive_map.html')



# -*- coding: utf-8 -*-

"""
This is a collection of helper functions that can be used within the tests.

This file is part of project DHNx (). It's copyrighted
by the contributors recorded in the version control history of the file,
available from its original location.

SPDX-License-Identifier: MIT
"""

import os

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal


def get_basic_path():
    """Returns the basic oemof path and creates it if necessary.
    The basic path is the '.oemof' folder in the $HOME directory.
    """
    basicpath = os.path.join(os.path.expanduser('~'), '.DHNx')
    if not os.path.isdir(basicpath):
        os.mkdir(basicpath)
    return basicpath


def extend_basic_path(subfolder):
    """Returns a path based on the basic oemof path and creates it if
     necessary. The subfolder is the name of the path extension.
    """
    extended_path = os.path.join(get_basic_path(), subfolder)
    if not os.path.isdir(extended_path):
        os.mkdir(extended_path)
    return extended_path


def get_all_file_paths(dir):
    # pylint: disable=unused-variable
    file_paths = []
    for dir_path, dir_names, file_names in os.walk(dir):
        for file_name in file_names:
            file_paths.append(os.path.join(dir_path, file_name))

    return file_paths


def check_if_csv_files_equal(csv_file_a, csv_file_b, check_arrays=False, **kwargs):
    r"""
    Compares two csv files.

    TODO: Check not working. Nothing happens, if check fails!

    Parameters
    ----------
    csv_file_a : str
        Path to first csv file

    csv_file_b : str
        Path to second csv file

    check_arrays : bool
        If True, only compares data, not index and columns
    """
    df1 = pd.read_csv(csv_file_a)
    df2 = pd.read_csv(csv_file_b)

    if check_arrays:
        np.array_equal(df1.values, df2.values, **kwargs)

    else:
        assert_frame_equal(df1, df2, **kwargs)


def check_if_csv_dirs_equal(dir_a, dir_b, **kwargs):
    r"""
    Compares the csv files in two directories and asserts that
    they are equal.

    The function asserts that:

    1. The file names of csv files found in the directories are the same.
    2. The file contents are the same.

    Parameters
    ----------
    dir_a : str
        Path to first directory containing csv files

    dir_b : str
        Path to second directory containing csv files

    """
    files_a = get_all_file_paths(dir_a)
    files_b = get_all_file_paths(dir_b)

    files_a = [file for file in files_a if file.split('.')[-1] == 'csv']
    files_b = [file for file in files_b if file.split('.')[-1] == 'csv']

    files_a.sort()
    files_b.sort()

    f_names_a = [os.path.split(f)[-1] for f in files_a]
    f_names_b = [os.path.split(f)[-1] for f in files_b]

    diff = list(set(f_names_a).symmetric_difference(set(f_names_b)))

    assert not diff, \
        f"Lists of filenames are not the same." \
        f" The diff is: {diff}"

    for file_a, file_b in zip(files_a, files_b):
        try:
            check_if_csv_files_equal(file_a, file_b, **kwargs)
        except AssertionError:
            diff.append([file_a, file_b])

    if diff:
        error_message = ''
        for pair in diff:
            short_name_a, short_name_b = (os.path.join(*f.split(os.sep)[-4:]) for f in pair)
            line = ' - ' + short_name_a + ' and ' + short_name_b + '\n'
            error_message += line

        raise AssertionError(f" The contents of these file are different:\n{error_message}")
